package it.volga;

import it.volga.ai.IntelligenceException;
import it.volga.ai.PortalAI;
import it.volga.ai.SokobanAI;
import it.volga.controller.AIController;
import it.volga.controller.InteractiveController;
import it.volga.model.Doom;
import it.volga.model.GameState;
import it.volga.view.EmptyView;
import it.volga.view.GUIView;
import it.volga.view.View;
import lombok.val;

import javax.swing.JOptionPane;
import javax.swing.SwingUtilities;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.Arrays;

/**
 * Автор: Евгений Грязнов
 */
public class App
{

    private static final String mapsDir = "maps/";
    private static final String console = "-console";

    public static void main( String[] args ) throws IOException, IntelligenceException
    {
        run(args);
//        ai(1);
    }

    private static void test(String[] args) throws IOException, IntelligenceException
    {
        if (args.length == 2)
        {
            ai(Integer.parseInt(args[1]));
        }
        else
            interactive(Integer.parseInt(args[0]));
    }

    private static void ai(int n) throws IOException, IntelligenceException
    {
        val modelForAI = new Doom("src/main/resources/maps/map" + n + ".txt");
        val modelForDemonstration = new Doom("src/main/resources/maps/map" + n + ".txt");
        val ai = new PortalAI(modelForAI);
        final View gui = new GUIView(modelForDemonstration);
        final it.volga.ai.Path solution = ai.solve();
        if (solution == null)
        {
            System.out.println("No solution :(");
        }
        else
        {
            System.out.println("Solution found: " + solution);
            val mainController = new AIController(gui, modelForDemonstration, solution);
            SwingUtilities.invokeLater(mainController::start);
        }
    }

    private static void run(String[] args) throws IOException
    {
        if (args.length == 0)
            System.out.println("Укажите файл с картой");
        else if (args.length == 1)
        {
            final String mapFilename = args[0];
            if (mapFilename.equals(console))
                System.out.println("Укажите файл с картой");
            else
                loadMap(mapFilename, false);
        }
        else
        {
            final String consoleArg = args[0];
            final String mapFilename = args[1];
            if (!consoleArg.equals(console))
                System.out.println("Первый аргумент должен быть -console");
            else
                loadMap(mapFilename, true);
        }
    }

    private static void loadMap(String mapFilename, boolean console) throws IOException
    {
        final Path pathToMap = Paths.get(mapsDir + mapFilename).toAbsolutePath();
        if (pathToMap.toFile().exists())
            ai(pathToMap, console);
        else
            System.out.printf("Файла %s не существует в папке maps/\n", mapFilename);
    }

    private static void ai(Path pathToMap, boolean console) throws IOException
    {
        val modelForAI = new Doom(pathToMap);
        val modelForDemonstration = new Doom(pathToMap);
        val ai = modelForAI.containsPortals() ? new PortalAI(modelForAI) : new SokobanAI(modelForAI);
        final View gui = console ? new EmptyView() : new GUIView(modelForDemonstration);
        try
        {
            final it.volga.ai.Path solution = ai.solve();
            if (solution == null)
                error("Не получилось найти решение!", console);
            else
            {
                System.out.println("Решение: " + solution);
                writeOutput(Integer.toString(solution.length()), solution.toString());
                val mainController = new AIController(gui, modelForDemonstration, solution);
                SwingUtilities.invokeLater(mainController::start);
            }
        }
        catch (IntelligenceException ie)
        {
            error("Не получилось найти решение, причина: " + ie.getMessage(), console);
        }
    }

    private static void error(String message, boolean console) throws IOException
    {
        System.out.println(message);
        if (!console)
            JOptionPane.showMessageDialog(null, message, "Ошибка", JOptionPane.ERROR_MESSAGE);
        writeOutput(message);
    }

    private static void writeOutput(String... lines) throws IOException
    {
        Files.write(Paths.get("output.txt"), Arrays.asList(lines));
    }

    private static void interactive(int mapNumber)
    {
        SwingUtilities.invokeLater(() ->
        {
            try
            {
                val model = new Doom("src/main/resources/maps/map" + mapNumber + ".txt");
                val gui = new GUIView(model);
                val mainController = new InteractiveController(gui, model);
                gui.addKeyListener(mainController);
                mainController.start();
            }
            catch (Exception e)
            {
                e.printStackTrace();
            }
        });
    }

}
