package it.volga.ai;

import it.volga.model.tiles.Door;
import lombok.val;
import org.jetbrains.annotations.NotNull;

import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Optional;
import java.util.stream.Collectors;

public class Path implements Iterable<Path.Node>, Comparable<Path>
{

    private final List<Node> nodes;

    Path()
    {
        this(new LinkedList<>());
    }

    public Path(final List<Node> nodes)
    {
        this.nodes = nodes;
    }

    public static Path backtrack(Vertex child)
    {
        return backtrack(new Path(), child);
    }

    private static Path backtrack(Path path, Vertex child)
    {
        if (child == null)
            return null;
        if (!child.hasParent())
        {
            return path;
        }
        else
        {
            val parent = child.parent();
            Move move = Move.resolve(parent.location(), child.location()).orElse(null);
            if (move != null)
            {
                path.nodes.add(0, new Node(move, child));
            }
            return backtrack(path, parent);
        }
    }

    public boolean isEmpty()
    {
        return nodes.isEmpty();
    }

    public Path.Node last()
    {
        return nodes.get(nodes.size() - 1);
    }

    public Path append(Path tail)
    {
        nodes.addAll(tail.nodes);
        return this;
    }

    @Override
    public String toString()
    {
        return nodes.stream()
                .map(Node::toString)
                .collect(Collectors.joining());
    }

    public void append(Vertex from, Vertex to)
    {
        val move = Move.resolve(from.location(), to.location())
                .orElseThrow(() -> new IllegalArgumentException("вершины должны быть инцидентны!"));
        append(new Node(move, to));
    }

    public void append(Path.Node node)
    {
        nodes.add(node);
    }

    public Path.Node[] nodes()
    {
        return nodes.toArray(new Path.Node[0]);
    }

    public int length()
    {
        return nodes.size();
    }

    public Path.Node get(final int offset)
    {
        final int index = offset < 0 ? length() + offset : offset;
        return nodes.get(index);
    }

    public Path copy()
    {
        val nodes = new LinkedList<Node>();
        for (Node node : this.nodes)
        {
            nodes.add(node.copy());
        }
        return new Path(nodes);
    }

    public Path concat(Path trail)
    {
        this.nodes.addAll(trail.nodes);
        return this;
    }

    public boolean containsDoors(List<Door> doors)
    {
        if (doors.isEmpty())
            return false;
        //
        for (Node node : this)
        {
            if (node.getVertex().onDoor())
            {
                val door = (Door) node.getVertex().tile();
                if (doors.contains(door))
                    return true;
            }
        }
        return false;
    }

    public boolean containsCrates()
    {
        for (Node node : this)
            if (node.getVertex().onCrate())
                return true;
        //
        return false;
    }

//    public Optional<Vertex> lastClosedDoor(final List<Key> keys)
//    {
//        Vertex lastClosedDoor = null;
//        for (Node n : nodes)
//        {
//            final Vertex v = n.vertex;
//            if (v.onDoor())
//            {
//                Door door = (Door) v.tile();
//                boolean closed = true;
//                for (Key key : keys)
//                    if (door.canBeOpenedBy(key))
//                    {
//                        closed = false;
//                        break;
//                    }
//                //
//                if (closed)
//                    lastClosedDoor = v;
//            }
//        }
//        return Optional.ofNullable(lastClosedDoor);
//    }
//
    @NotNull
    @Override
    public Iterator<Node> iterator()
    {
        return nodes.iterator();
    }

    @Override
    public int compareTo(@NotNull Path o)
    {
//        return Comparator.comparing(Path::totalDamage).thenComparing(Path::length).compare(this, o);
        return 0; // TODO FIx
    }

    public static class Node
    {
        private Move move;
        private final Vertex vertex;

        public Node(Move move, Vertex vertex)
        {
            this.move = move;
            this.vertex = vertex;
        }

        public void reverseMove()
        {
            this.move = move.reverse();
        }

        public Move getMove()
        {
            return move;
        }

        public Vertex getVertex()
        {
            return vertex;
        }

        public Node copy()
        {
            return new Node(move, vertex);
        }


        @Override
        public String toString()
        {
            return Character.toString(move.letter());
        }
    }


}
