package it.volga.ai;

import it.volga.model.Doom;
import it.volga.model.Location;
import it.volga.model.tiles.Crate;
import it.volga.model.tiles.Floor;
import lombok.val;

import java.util.LinkedList;
import java.util.List;
import java.util.SplittableRandom;

public class SokobanAI extends AI
{

    public SokobanAI(Doom model)
    {
        super(model);
    }

    @Override
    public Path solve() throws IntelligenceException
    {
        System.out.println("sokoban AI");
        final Path shortest = graph.shortestPath(graph.spawn(), graph.exit());
        if (shortest == null || shortest.length() == 0)
        {
            throw new IntelligenceException("Нет простого пути до выхода или мы уже там, карта неверна!");
        }
        return solveRec(graph.spawn(), shortest);
    }

    private Path solveRec(Vertex player, final Path road)
    {
        if (empty(road))
            return road;
        //
        if (road.length() == 1 && road.get(0).getVertex().onExit())
            return road;
        //
        final List<Path.Node> cleanPrefix = new LinkedList<>();
        Path.Node firstCrate = null;
        final List<Path.Node> tail = new LinkedList<>();
        for (Path.Node node : road)
        {
            if (firstCrate == null)
                if (onCrate(node.getVertex()))
                    firstCrate = node;
                else
                    cleanPrefix.add(node);
            else
                tail.add(node);
        }
        // путь не содержит ящиков
        if (firstCrate == null)
            return road;
        // нашли первый ящик, идём до него
        for (Path.Node prefix : cleanPrefix)
        {
            model.step(prefix.getVertex().location());
            player = prefix.getVertex();
        }
        // выбираем что с ним делать, пробуем обойти
        final Path sidestep = sidestep(player, firstCrate.getVertex());
        // пробуем убрать с дороги
        final Path moveAway = remove(player, firstCrate.getVertex());
        // стратегия
        if (empty(sidestep) && empty(moveAway))
        {
            // ничего не сработало, просто двигаем ящик
            if (moveCrate(firstCrate))
            {
                tail.add(0, firstCrate);
                final Path theRest = solveRec(firstCrate.getVertex(), new Path(tail));
                // нет пути
                if (theRest == null)
                    return null;
                // выход есть!
                return new Path(cleanPrefix).append(theRest);
            }
            else
                // дальше двигать нельзя, пути нет
                return null;
        }
        else if (populated(moveAway) && (empty(sidestep) || moveAway.length() < sidestep.length()))
        {
            // стратегия "убрать ящик" сработала и лучше чем "обойти"
            final Path theRest = solveRec(firstCrate.getVertex(), new Path(tail));
            // нет пути
            if (theRest == null)
                return null;
            return new Path(cleanPrefix).append(moveAway).append(theRest);
        }
        else if (populated(sidestep))
        {
            // выбираем стратегию "обойти"
            final Path theRest = solveRec(player, sidestep);
            // нет пути
            if (theRest == null)
                return null;
            return new Path(cleanPrefix).append(theRest);
        }
        else
            return null;
    }

    private boolean moveCrate(Path.Node to)
    {
        if (model.step(to.getVertex().location()))
        {
            to.getVertex().setTile(new Floor());
            for (Move value : Move.values())
            {
                Vertex u = to.getVertex().neighborInDirection(value);
                if (u != null)
                {
                    u.setTile(model.tile(u.location()));
                }
            }
            return true;
        }
        return false;
    }

    private boolean isHorizontal(Location playerAt, Location crateAt)
    {
        return playerAt.row() == crateAt.row();
    }

    private Vertex[] horizontal(Vertex crate)
    {
        return new Vertex[]{crate.neighborInDirection(Move.LEFT), crate.neighborInDirection(Move.RIGHT)};
    }

    private Vertex[] vertical(Vertex crate)
    {
        return new Vertex[]{crate.neighborInDirection(Move.UP), crate.neighborInDirection(Move.DOWN)};
    }

    // стоим перед ящиком, пытаемся убрать с дороги
    private Path remove(Vertex playerAt, Vertex firstCrate)
    {
        final Vertex[] disposition;
        if (isHorizontal(playerAt.location(), firstCrate.location()))
            disposition = vertical(firstCrate);
        else
            disposition = horizontal(firstCrate);
        //
        if (isWalkable(disposition[0]) && isWalkable(disposition[1]))
        {
            final Path toFirst = graph.walkableShortestPath(playerAt, disposition[0]);
            if (empty(toFirst))
            {
                final Path result = Path.backtrack(disposition[1]);
                if (empty(result))
                {
                    return null;
                }
                val move = Move.resolve(disposition[1].location(), firstCrate.location()).orElseThrow(() -> new IllegalStateException("asdf"));
                final Path.Node last = new Path.Node(move, firstCrate);
                result.append(last);
                return result;
            }
            else
            {
                val move = Move.resolve(disposition[0].location(), firstCrate.location()).orElseThrow(() -> new IllegalStateException("asdf"));
                final Path.Node last = new Path.Node(move, firstCrate);
                toFirst.append(last);
                return toFirst;
            }
        }
        else
            return null;
    }

    private boolean isWalkable(final Vertex v)
    {
        return v != null && model.tile(v.location()).isWalkable();
    }

    // стоим перед ящиком, пытаемся обойти
    private Path sidestep(Vertex playerAt, Vertex firstCrate)
    {
        final List<Vertex> ns = firstCrate.neighbours();
        firstCrate.isolate();
        final Path result = graph.shortestPath(playerAt, graph.exit());
        firstCrate.setNeighbours(ns);
        return result;
    }

    private boolean populated(Path path)
    {
        return path != null && !path.isEmpty();
    }

    private boolean empty(Path path)
    {
        return path == null || path.isEmpty();
    }

    private boolean onCrate(Vertex v)
    {
        return model.tile(v.location()) instanceof Crate;
    }

}
