package it.volga.ai;

import it.volga.model.Location;
import it.volga.model.tiles.Exit;
import it.volga.model.tiles.Spawn;
import it.volga.model.tiles.Tile;
import lombok.val;

import java.util.Deque;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.function.Consumer;
import java.util.stream.Collectors;

public class Graph
{

    private final List<Vertex> vertices = new LinkedList<>();
    private Vertex spawn;
    private Vertex exit;

    public Graph()
    {
    }

    public Vertex spawn()
    {
        return spawn;
    }

    public List<Vertex> vertices()
    {
        return vertices;
    }

    public Vertex exit()
    {
        return exit;
    }

    public Vertex addVertex(Tile tile, Location location)
    {
        val result = new Vertex(tile, location);
        vertices.add(result);
        if (tile instanceof Spawn)
        {
            this.spawn = result;
        }
        else if (tile instanceof Exit)
        {
            this.exit = result;
        }
        return result;
    }

    public int order()
    {
        return vertices.size();
    }

    public void chain(Vertex a, Vertex b)
    {
        a.connectTo(b);
        b.connectTo(a);
    }

    private void clear()
    {
        for (Vertex vertex : vertices)
            vertex.clear();
    }

    /**
     * Высчитывает вес всех рёбер, входящих в вершину to.
     * В графе карты все вершины имеют входящие рёбра с одинаковым весом.
     * Вес ребра, входящего в вершину с ячейкой огня равен урону от этой ячейки
     * Вес ребер, входящих в остальные вершины равен 1.
     * @param to вершина, вес входящих ребер которой нужно посчитать
     * @return вес входящих рёбер вершины to
     */
    // public int weight(Vertex to)
    // {
    //     if (to.onFire())
    //     {
    //         val fireTile = (Fire) to.tile();
    //         return fireTile.damage();
    //     }
    //     else
    //         return 1;
    // }

    private void walkableBFS(Vertex root)
    {
        final Deque<Vertex> forVisit = new LinkedList<>();
        forVisit.add(root);
        root.mark();
        while (!forVisit.isEmpty())
        {
            final Vertex u = forVisit.removeFirst();
            for (Vertex neighbour : u.walkableNeighbors())
            {
                if (!neighbour.marked())
                {
                    neighbour.mark();
                    neighbour.parent(u);
                    forVisit.add(neighbour);
                }
            }
        }
    }

    private void bfs(Vertex root)
    {
        final Deque<Vertex> forVisit = new LinkedList<>();
        forVisit.add(root);
        root.mark();
        while (!forVisit.isEmpty())
        {
            final Vertex u = forVisit.removeFirst();
            for (Vertex neighbour : u.neighbours())
            {
                if (!neighbour.marked())
                {
                    neighbour.mark();
                    neighbour.parent(u);
                    forVisit.add(neighbour);
                }
            }
        }
    }

//    public List<Path> safePaths(Vertex from, Vertex to)
//    {
//        Path safest = null;
//        List<Path> safestPaths = new LinkedList<>();
//        for (Path path : paths(from, to))
//            if (safest == null || path.compareTo(safest) < 0)
//            {
//                safest = path;
//                safestPaths = new LinkedList<>();
//                safestPaths.add(safest);
//            }
//            else if (path.compareTo(safest) == 0)
//            {
//                safestPaths.add(path);
//            }
//        //
//        return safestPaths;
//    }
//

    public List<Path> paths(Vertex from, Vertex to)
    {
        val allPaths = new LinkedList<Path>();
        val visited = new HashSet<Vertex>();
        paths(from, to, new Path(), visited, allPaths);
        return allPaths;
    }

    private void paths(Vertex from, Vertex to, Path traveled, HashSet<Vertex> visited, List<Path> accumulator)
    {
        if (from.equals(to))
            accumulator.add(traveled);
        else if (!visited.contains(from))
        {
            if (from.neighbours().size() == 1)
            {
                val onlyNeighbour = from.neighbours().get(0);
                traveled.append(from, onlyNeighbour);
                visited.add(from);
                paths(onlyNeighbour, to, traveled, visited, accumulator);
            }
            else
            {
                for (Vertex neighbour : from.neighbours())
                {
                    val traveledCopy = traveled.copy();
                    traveledCopy.append(from, neighbour);
                    val visitedCopy =  (HashSet<Vertex>) visited.clone();
                    visitedCopy.add(from);
                    paths(neighbour, to, traveledCopy, visitedCopy, accumulator);
                }
            }
        }
    }

    public boolean connected(Vertex from, Vertex to)
    {
        return from.neighbours().contains(to);
    }

    public void dfs(Vertex root, Consumer<Vertex> visit)
    {
        clear();
        dfsRec(root, visit);
    }

    private void dfsRec(Vertex root, Consumer<Vertex> visit)
    {
        if (!root.marked())
        {
            root.mark();
            visit.accept(root);
            for (Vertex neighbour : root.neighbours())
            {
                neighbour.parent(root);
                dfsRec(neighbour, visit);
            }
        }
    }

    private void clean()
    {
        vertices.forEach(Vertex::clear);
    }

    public Path walkableShortestPath(Vertex head, Vertex tail)
    {
        clear();
        walkableBFS(head);
        return Path.backtrack(tail);
    }

    public Path shortestPath(Vertex head, Vertex tail)
    {
        clear();
        bfs(head);
        return Path.backtrack(tail);
    }

    @Override
    public String toString()
    {
        return vertices.stream().map(Vertex::toString).collect(Collectors.joining("\n"));
    }
}
