package it.volga.ai;

import it.volga.model.Location;
import it.volga.model.tiles.Crate;
import it.volga.model.tiles.Door;
import it.volga.model.tiles.Exit;
import it.volga.model.tiles.Tile;
import lombok.val;
import sun.plugin.dom.exception.InvalidStateException;

import java.util.LinkedList;
import java.util.List;
import java.util.stream.Collectors;

public class Vertex
{
    private static int NEXT_ID = 0;

    private List<Vertex> neighbours = new LinkedList<>();
    private final int id;
    private Tile tile;
    private final Location location;
    private boolean marked;
    private Vertex parent;

    Vertex(Tile tile, Location location)
    {
        this.id = NEXT_ID++;
        this.tile = tile;
        this.location = location;
    }

    public void clear()
    {
        unmark();
        this.parent = null;
    }

    public void unmark()
    {
        this.marked = false;
    }

    public void mark()
    {
        this.marked = true;
    }

    public boolean marked()
    {
        return marked;
    }

    public List<Vertex> walkableNeighbors()
    {
        return neighbours.stream()
                .filter(v -> v.tile.isWalkable())
                .collect(Collectors.toList());
    }

    public void setTile(Tile tile)
    {
        this.tile = tile;
    }

    public List<Vertex> neighbours()
    {
        return neighbours;
    }

    public void isolate()
    {
        for (Vertex neighbour : neighbours)
            neighbour.dropEdgeTo(this);
        neighbours = new LinkedList<>();
    }

    public void dropEdgeTo(Vertex another)
    {
        clear();
        neighbours.remove(another);
    }

    public void connectTo(Vertex another)
    {
        neighbours.add(another);
    }

    public int id()
    {
        return id;
    }

    @Override
    public String toString()
    {
        val neighbourIds = neighbours.stream()
                .map(v -> Integer.toString(v.id))
                .collect(Collectors.joining(","));
        return String.format("%d -> %s [%s] %s", id, neighbourIds, tile, location);
    }

    public Tile tile()
    {
        return tile;
    }

    public Vertex neighborInDirection(Move direction)
    {
        for (Vertex neighbour : neighbours)
            if (Move.resolve(location, neighbour.location).orElseThrow(() -> new InvalidStateException("asdf")) == direction)
                return neighbour;
//        throw new InvalidStateException("не нашёл соседа в таком направлении");
        return null;
    }

    public void setNeighbours(List<Vertex> neighbours)
    {
        this.neighbours = neighbours;
    }

    public boolean onExit()
    {
        return tile instanceof Exit;
    }

    public boolean onCrate()
    {
        return tile instanceof Crate;
    }

    public boolean onDoor()
    {
        return tile instanceof Door;
    }

    public Location location()
    {
        return location;
    }

    public void parent(Vertex parent)
    {
        this.parent = parent;
    }

    public boolean hasParent()
    {
        return parent != null;
    }

    public Vertex parent()
    {
        return parent;
    }

}
