package it.volga.model;

import it.volga.model.tiles.*;
import lombok.val;

import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;
import java.util.stream.Collectors;

/**
 * Текстовая модель карты
 */
public class Doom
{

    private Tile[][] grid;
    private final Rook rook;
    private Location spawnAt;
    private Location exitAt;
    private GameState state;
    private List<Location> crateLocations;
    private List<Location> portalLocations;

    public Doom(String filename) throws IOException
    {
        this(Paths.get(filename));
    }

    public Doom(Path pathToMap) throws IOException
    {
        read(pathToMap);
        this.rook = new Rook();
        this.rook.move(spawnAt);
        this.state = GameState.NEW;
    }

    private void read(Path pathToMap) throws IOException
    {
        val lines = Files.lines(pathToMap).collect(Collectors.toList());
        val dimensions = lines.get(0).split(" ");
        final int nOfRows = Integer.parseInt(dimensions[0]);
        final int nOfColumns = Integer.parseInt(dimensions[1]);
        this.grid = new Tile[nOfRows][nOfColumns];
        this.crateLocations = new ArrayList<>();
        this.portalLocations = new LinkedList<>();
        int row = 0;
        int col = 0;
        for (int i = 1; i < lines.size(); i++)
        {
            val line = lines.get(i);
            for (int j = 0; j < line.length(); j++)
            {
                final char symbol = line.charAt(j);
                if (symbol == ' ' || symbol == '\n')
                    continue;
                val nextTile = TileResolver
                        .resolve(symbol)
                        .orElseThrow(() -> new IllegalArgumentException("В файле карты встретился неизвестный символ " + symbol));
                grid[row][col] = nextTile instanceof Crate ? new Floor() : nextTile;
                val current = Location.of(row, col);
                switch (symbol)
                {
                    case 'S':
                        spawnAt = current;
                        break;
                    case 'Q':
                        exitAt = current;
                        break;
                    case 'K':
                        crateLocations.add(current);
                        break;
                    case 'F':
                    case 'G':
                    case 'H':
                    case 'I':
                    case 'J':
                    case 'f':
                    case 'g':
                    case 'h':
                    case 'i':
                    case 'j':
                        portalLocations.add(current);
                        break;
                }
                col += 1;
            }
            col = 0;
            row += 1;
        }
        if (spawnAt == null)
            throw new IllegalArgumentException("На карте нет точки спавна!");
        if (exitAt == null)
            throw new IllegalArgumentException("На карте нет точки выхода!");
    }

    public Location spawnAt()
    {
        return spawnAt;
    }

    public Tile tile(Location at)
    {
        return tile(at.row(), at.col());
    }

    public boolean step(Location location)
    {
        val tile = tile(location);
        if (tile.tryToStep(this, location))
        {
            this.state = computeState();
            return true;
        }
        return false;
    }

    private GameState computeState()
    {
        if (rook.at().equals(this.exitAt()))
            return GameState.WIN;
        else
            return GameState.PLAYING;
    }

    public Location exitAt()
    {
        return exitAt;
    }

    public GameState state()
    {
        return state;
    }

    public Tile tile(final int row, final int col)
    {
        if (row < 0 || row >= grid.length || col < 0 || col >= grid[0].length)
            throw new IllegalArgumentException(String.format("Неверные координаты: %d, %d", row, col));
        //
        val currentLoc = Location.of(row, col);
        for (Location crateLoc : crateLocations)
            if (currentLoc.equals(crateLoc))
                return new Crate();
        return grid[row][col];
    }

    public void setTile(Location at, Tile tile)
    {
        grid[at.row()][at.col()] = tile;
    }

    public int rows()
    {
        return grid.length;
    }

    public int cols()
    {
        return grid[0].length;
    }

    public Rook player()
    {
        return rook;
    }

    public void teleport(Portal dual)
    {
        for (Location portalLocation : portalLocations)
        {
            val portal = (Portal) tile(portalLocation);
            if (portal.equals(dual))
            {
                System.out.println("Телепорт через портал " + portal.symbol());
                player().move(portalLocation);
                break;
            }
        }
    }

    public boolean tryMoveCrateAt(Location crateAt)
    {
        final Location delta = rook.at().delta(crateAt);
        final Location movedCratePosition = crateAt.translate(delta);
        if (tile(movedCratePosition).tryMoveCrateTo())
        {
            for (int i = 0; i < crateLocations.size(); i++)
                if (crateLocations.get(i).equals(crateAt))
                {
                    crateLocations.set(i, movedCratePosition);
                    break;
                }
            return true;
        }
        return false;
    }

}
