package it.volga.view;

import it.volga.model.Doom;
import it.volga.model.Location;
import it.volga.view.stats.Statistics;
import it.volga.view.stats.Steps;
import lombok.val;

import javax.swing.*;
import java.awt.*;

/**
 * Отображение игры в виде GUI
 */
public class GUIView extends JFrame implements View
{
    private static final Point gridBase = Point.of(0, 20);
    private static final Color BACKGROUND = Color.GRAY;

    private final Doom model;
    private int spriteWidth;
    private int spriteHeight;
    private final Canvas canvas;
    private Statistics[] stats;

    public GUIView(final Doom model)
    {
        this.model = model;
        setupStatistics();
        calculateSpriteDimensions();
        setupWindowSize();
        setTitle("Doom by Evgeniy Gryaznov");
        setResizable(false);
        setLocationRelativeTo(null);
        setDefaultCloseOperation(EXIT_ON_CLOSE);
        Sprite.PLAYER.setPicture(Sprite.PLAYER.picture().getScaledInstance(spriteWidth, spriteHeight, Image.SCALE_SMOOTH));
        this.canvas = new Canvas();
        this.canvas.setBackground(BACKGROUND);
        add(this.canvas);
        setVisible(false);
    }

    private void setupStatistics()
    {
        this.stats = new Statistics[]{
                new Steps(this.model.player(), BACKGROUND),
        };
    }

    private void calculateSpriteDimensions()
    {
        val screen = Toolkit.getDefaultToolkit().getScreenSize();
        final int maxWidth = screen.width / model.cols();
        final int maxHeight = (screen.height - 48) / model.rows();
        final int spriteDims = Math.min(maxHeight, maxWidth);
        this.spriteWidth  = spriteDims;
        this.spriteHeight = spriteDims;
    }

    private void setupWindowSize()
    {
        final int minWidth = 300;
        final int windowHeight = this.spriteHeight * model.rows() + 48;
        final int windowWidth = Math.max(minWidth, this.spriteWidth * model.cols());
        setSize(windowWidth, windowHeight);
    }

    @Override
    public void update()
    {
        canvas.update(canvas.getGraphics());
    }

    @Override
    public void renderTileAt(Location location)
    {
        canvas.renderTileAt(location);
    }

    @Override
    public void present()
    {
        setVisible(true);
    }

    @Override
    public void conceal()
    {
        setVisible(false);
    }

    private class Canvas extends JPanel
    {

        @Override
        public void paint(Graphics g)
        {
            super.paint(g);
            val g2d = (Graphics2D)g;
            g2d.setColor(Color.BLACK);
            renderStats(g2d);
            renderGrid(g2d);
            renderPlayer(g2d);
        }

        @Override
        public void update(Graphics g)
        {
            val g2d = (Graphics2D)g;
            g2d.setColor(Color.BLACK);
            switch (model.state())
            {
                case WIN:
                    drawSign(g2d, "ПОБЕДА!");
                    break;
                case PLAYING:
                    renderStats(g2d);
                    renderPlayer(g2d);
                    break;
            }
        }

        private void renderStats(Graphics2D g2d)
        {
            final int offsetX = 20;
            final Point offset = Point.of(offsetX, 0);
            final int baseX = 1;
            Point base = Point.of(baseX, Statistics.LETTER_HEIGHT);
            for (Statistics stat : stats)
            {
                base = stat.render(g2d, base);
                base.move(offset);
            }
        }

        public void renderTileAt(Location location)
        {
            final Graphics2D g2d = (Graphics2D) getGraphics();
            val tile = model.tile(location.row(), location.col());
            val sprite = Sprite.resolve(tile).orElseThrow(() -> new IllegalStateException("потерялся спрайт"));
            val point = locationToPoint(location);
            drawSpriteAt(g2d, sprite, point);
        }

        private void drawSign(final Graphics2D g2d, final String sign)
        {
            final int middleX = getWidth() / 2;
            final int middleY = getHeight() / 2;
            g2d.setColor(Color.RED);
            g2d.setFont(g2d.getFont().deriveFont(Font.BOLD, 75));
            g2d.drawString(sign, middleX - 220, middleY - 15);
        }

        private void renderPlayer(final Graphics2D g2d)
        {
            val player = model.player();
            val playerSprite = Sprite.PLAYER;
            val drawAt = locationToPoint(player.at());
            drawSpriteAt(g2d, playerSprite, drawAt);
        }

        private Point locationToPoint(final Location l)
        {
            final int x = spriteWidth * (l.col()) + gridBase.axis();
            final int y = spriteHeight * (l.row()) + gridBase.ordinate();
            return Point.of(x, y);
        }

        private void renderGrid(final Graphics2D g2d)
        {
            int x = gridBase.axis();
            int y = gridBase.ordinate();
            for (int i = 0; i < model.rows(); i++)
            {
                for (int j = 0; j < model.cols(); j++)
                {
                    val sprite = tileToSprite(i, j);
                    val picture = sprite.picture();
                    if (spriteWidth != picture.getWidth(null) || spriteHeight != picture.getHeight(null))
                        sprite.setPicture(picture.getScaledInstance(spriteWidth, spriteHeight, Image.SCALE_SMOOTH));
                    drawSpriteAt(g2d, sprite, x, y);
                    x += spriteWidth;
                }
                x = gridBase.axis();
                y += spriteHeight;
            }
        }

        private Sprite tileToSprite(final int tileRow, final int tileCol)
        {
            val tile = model.tile(tileRow, tileCol);
            return Sprite.resolve(tile).orElseThrow(() -> new IllegalStateException("нет спрайта для ячейки " + tileRow + " " + tileCol));
        }

        private void drawSpriteAt(final Graphics2D g2d, final Sprite sprite, final Point at)
        {
            drawSpriteAt(g2d, sprite, at.axis(), at.ordinate());
        }

        private void drawSpriteAt(final Graphics2D g2d, final Sprite sprite, final int x, final int y)
        {
            g2d.drawImage(sprite.picture(), x, y, null);
        }

    }
}
